import io
import os
import subprocess
import threading
from _winapi import CREATE_NO_WINDOW
from pathlib import Path
from shutil import rmtree
from subprocess import Popen

import chardet
import requests

from launcher import config
from launcher.constants import download_git_url, download_python_url, get_new_version_url, upgrade_script_name, \
    mirror_list
from launcher.file.download import init_download, download_file
from launcher.file.extract import extract_7z_file
from launcher.file.utils import ensure_directory_exists
from launcher.log import logger
from launcher.sys_config import sys_config
from launcher.webview.events import custom_event

command_list = {
    "download_git": lambda: init_download("git", download_git_url, os.getcwd()),
    "download_python": lambda: init_download("python", download_python_url, os.getcwd()),
    "lfs": "git\\bin\\git lfs install",
    "ensurepip": "python\\python -m ensurepip --default-pip",
    "clone": "git\\bin\\git -c lfs.concurrenttransfers=100 clone https://git.zhaozuohong.vip/mower-ng/mower-ng.git --branch slow",
    "fetch": lambda: f"..\\git\\bin\\git fetch origin {config.conf.branch} --progress",
    "switch": lambda: f"..\\git\\bin\\git -c lfs.concurrenttransfers=100 switch -f {config.conf.branch} --progress",
    "reset": lambda: f"..\\git\\bin\\git -c lfs.concurrenttransfers=200 reset --hard origin/{config.conf.branch}",
    "pip_install": lambda: f"..\\python\\Scripts\\pip install --no-cache-dir -i {mirror_list[config.conf.mirror]} -r requirements.txt --no-warn-script-location",
    "webview": "start ..\\python\\pythonw webview_ui.py",
    "manager": "start ..\\python\\pythonw manager.py",
}


def detect_encoding(stream):
    # 读取一部分数据来检测编码
    raw_data = stream.read(4096)
    result = chardet.detect(raw_data)
    stream.seek(0)  # 将流指针重置到开头
    return result['encoding']


def read_stream(stream, log_func):
    detected_encoding = detect_encoding(stream)
    text_io = io.TextIOWrapper(stream, encoding=detected_encoding, errors='replace')
    try:
        for line in iter(text_io.readline, ''):
            text = line.rstrip('\n')
            custom_event(text.strip() + "\n")
    finally:
        text_io.close()


class Api:

    def load_config(self):
        logger.info("读取配置文件")
        return config.conf.model_dump()

    def save_config(self, conf):
        logger.info(f"更新配置文件{conf}")
        config.conf = config.Conf(**conf)
        config.save_conf()

    def get_version(self):
        return sys_config.get("version")

    def get_new_version(self):
        logger.info("获取最新版本号")
        response = requests.get(get_new_version_url)
        return response.json()

    # 更新启动器本身
    def update_self(self, download_url):
        logger.info(f"开始更新启动器 {download_url}")
        file_name = os.path.basename(download_url)
        file_name = "launcher.7z"
        current_path = os.getcwd()
        download_tmp_folder = os.path.join(current_path, "download_tmp")
        # 确保 download_tmp 文件夹存在
        ensure_directory_exists(download_tmp_folder)
        if not download_file("launcher", download_url, download_tmp_folder):
            return "下载新版本失败"

        download_path = os.path.join(download_tmp_folder, file_name)
        if not extract_7z_file("launcher", download_path, download_tmp_folder, True):
            return "解压新版本失败"

        exe_path = os.path.join(current_path, "launcher.exe")
        folder_path = os.path.join(current_path, "_internal")
        new_exe_path = os.path.join(download_tmp_folder, "launcher", "launcher.exe")
        new_folder_path = os.path.join(download_tmp_folder, "launcher", "_internal")
        script_path = os.path.join(download_tmp_folder, upgrade_script_name)
        with open(script_path, "w") as b:
            temp_list = "@echo off\n"
            temp_list += "timeout /t 3 /nobreak\n"  # 等待进程退出
            temp_list += f"rmdir {folder_path}\n"  # 删除_internal
            temp_list += f"del {exe_path}\n"  # 删除exe
            temp_list += f"xcopy /e /i /y {new_folder_path} {folder_path}\n"
            temp_list += f"move {new_exe_path} {exe_path}\n"
            temp_list += "timeout /t 1 /nobreak\n"  # 等待操作完成
            temp_list += f"start {exe_path}\n"  # 启动新程序
            temp_list += "exit"
            b.write(temp_list)
        # 不显示cmd窗口
        Popen([script_path], creationflags=CREATE_NO_WINDOW)
        os._exit(0)

    def rm_site_packages(self):
        site_packages_path = Path("./python/Lib/site-packages")
        if site_packages_path.exists():
            rmtree(site_packages_path)
            return "site-packages目录移除成功"
        return "python\\Lib\\site-packages目录不存在"

    def rm_python_scripts(self):
        python_scripts_path = Path("./python/Scripts")
        if python_scripts_path.exists():
            rmtree(python_scripts_path)
            return "Scripts目录移除成功"
        return "python\\Scripts目录不存在"

    def run(self, command, cwd=None):
        command = command_list[command]
        if callable(command):
            command = command()
            if callable(command):
                return "success" if command() else "failed"
        custom_event(command + "\n")
        try:
            with subprocess.Popen(
                    command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, cwd=cwd, bufsize=0,
                    universal_newlines=False
            ) as p:
                stdout_thread = threading.Thread(target=read_stream, args=(p.stdout, logger.info))
                stderr_thread = threading.Thread(target=read_stream, args=(p.stderr, logger.info))

                stdout_thread.start()
                stderr_thread.start()

                stdout_thread.join()
                stderr_thread.join()
            if p.returncode == 0:
                return "success"
        except Exception as e:
            logger.exception(e)
            custom_event(str(e))
        return "failed"
